/*************************************************************************
 * mmol2mol.cpp - Program for conversion of mmol files to common         *
 *                file formats.                                          *
 *                                                                       *
 * Copyright (C) 2014-2015 by Laboratory of Medicinal Chemistry of       *
 * Department of Chemistry, Lomonosov Moscow State University            *
 *                                                                       *
 * Programmed by Oleg. I. Titov                                          *
 *                                                                       *
 * This file is part of the Electrostatic Tools project.                 *
 * For more information, see <http://qsar.chem.msu.ru/elec_tools/>       *
 *                                                                       *
 * Based on work:                                                        *
 * O.I. Titov, D.A. Shulga, V.A. Palyulin, N.S. Zefirov                  *
 * Description of halogen bonding on the basis of multicenter multipole  *
 * expansion // Doklady Chemistry, 2013, vol. 450, p. 139                *
 *                                                                       *
 * This program is free software; you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation version 2 of the License.                *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 ************************************************************************/

#include <fstream>
#include <iomanip>
#include <cstring>
#include <vector>
#include <cctype>

#include <boost/program_options.hpp>

#include <openbabel/obconversion.h>

#include <Eigen/Dense>

#include "mep_fitter.h"
#include "molecule.h"
#include "comment_skipper.h"
#include "elec_tools_errors.h"
#include "elec_tools_helpers.h"
#include "config.h"


namespace po = boost::program_options;

int main(int argc, char const *const *argv)
{
  po::options_description all_opts;

  po::options_description generic_opts("MMol to common format converter v " + project_version + "\n\n"
      "Usage:\n"
      "mmol2mol <options>\n"
      "mmol2mol <in_mol> <out_mol> [options]\n\n"
      "Options available");
  add_generic_options(generic_opts);

  po::options_description input_opts("Input control");
  input_opts.add_options()
                ("input,I", po::value<std::string>()->required(), "<required> input molecule file");


  po::options_description output_opts("Output control");
  output_opts.add_options()
                 ("output,O", po::value<std::string>(), "output molecule file. Default: <input>.mol2")
                 ("filetype,t", po::value<std::string>()->default_value("GUESS"), "output molecule file type extension")
                 ("force-output,f", po::bool_switch()->default_value(false), "owerwrite existing files");

  po::options_description conv_opts("Conversion control");
  conv_opts.add_options()
               ("radius,r", po::value<double>()->default_value(0.1, "0.1"), "charge cluster radius in Angstroms")
               ("ignore-dipole,d", po::bool_switch()->default_value(false), "do not include dipole in charge cluster")
               ("ignore-quadrupole,q", po::bool_switch()->default_value(false), "do not include quadrupole in charge cluster")
               ("mcc-mask,M", po::value<std::vector<std::string> >(), "SMARTS mask of atoms which multipoles are to be converted to MCC (multiple keys are posible)");

  all_opts.add(generic_opts);
  all_opts.add(input_opts);
  all_opts.add(output_opts);
  all_opts.add(conv_opts);

  po::positional_options_description positional;
  positional.add("input", 1);
  positional.add("output", 1);

  po::variables_map vm;
  try
  {
    po::store(po::command_line_parser(argc, argv).options(all_opts).positional(positional).run(), vm);
    if (process_generic_opts(vm, all_opts, "MMol to common format converter")) return 0;
    po::notify(vm);
  } catch (const boost::program_options::required_option& e)
  {
    std::cout << all_opts << std::endl;
    std::cerr << "Error: " << e.what() << std::endl;
    return ELEC_TOOLS_BAD_OPTION;
  } catch (const boost::program_options::unknown_option& e)
  {
    std::cout << all_opts << std::endl;
    std::cerr << "Error: " << e.what() << std::endl;
    return ELEC_TOOLS_BAD_OPTION;
  }

  try
  {
    std::string filetype = vm["filetype"].as<std::string>();
    std::transform(filetype.begin(), filetype.end(), filetype.begin(), tolower);
    if (filetype == "guess")
    {
      if ( ! vm.count("output") )
      {
        std::cerr << "Format guess requested but no filename specified. Using MOL2 format." << std::endl;
        filetype = "mol2";
      } else {
        std::string t = vm["output"].as<std::string>();
        filetype = t.substr(t.find_last_of(".")+1);
      }
    }

    std::string outname;
    if ( ! vm.count("output") )
      outname = add_filename_suffix(vm["input"].as<std::string>(), "mcc", filetype);
    else outname = vm["output"].as<std::string>();

    if (!vm["force-output"].as< bool >())
    {
      if (check_output_exists(outname))
      {
        std::cerr << "Error: output molecule file exists. Aborting." << std::endl;
        return ELEC_TOOLS_OUTPUT_EXISTS;
      }
    }

    std::ostream* outstrm;

    int err = open_output(&outstrm, outname); if(err) return err;

    double radius = vm["radius"].as< double >();
    if (radius > 1.0)
      std::cout << "Your MCC radius is larger than 1.0 Angstroms. This is strange since it is expected to be as small as possible." << std::endl;
    radius /= bohr;

    bool strip_dipole = vm["ignore-dipole"].as< bool >();
    bool strip_quadrupole = vm["ignore-quadrupole"].as< bool >();

    std::ifstream input(vm["input"].as< std::string >().c_str());
    if (input.bad())
    {
      std::cerr << "Error opening input molecule file: \"" << outname << "\". Aborting." << std::endl;
      return ELEC_TOOLS_NO_INPUT_FILE;
    }

    GeneralMultipoledMolecule mol;
    CommentSkipper skipper;
    std::istringstream ifile(skipper.skip(input));
    if (!mol.readMe(ifile).empty())
    {
      std::cerr << "Failure reading input molecule. Aborting." << std::endl;
      return ELEC_TOOLS_MOL_READ_FAILURE;
    }
    input.close();

    OpenBabel::OBConversion conv;
    if (!conv.SetOutFormat(filetype.c_str()))
    {
      std::cerr << "Error setting output format. Aborting." << std::endl;
      return ELEC_TOOLS_BAD_MOL_FORMAT;
    }
    conv.SetOutStream(outstrm);

    if (strip_dipole && strip_quadrupole) // no need for MCC
    {
      if (!conv.Write(&mol))
      {
        std::cerr << "Error writing final molecule. Aborting." << std::endl;
        return ELEC_TOOLS_OUTPUT_FAILURE;
      } else return 0;
    }

    std::vector< std::string > mcc_patterns;
    if (vm.count("mcc-mask"))
      mcc_patterns = vm["mcc-mask"].as< std::vector< std::string > >();
    std::vector< int > mcc_matches = get_unique_SMARTS_matches(mcc_patterns, mol);

    std::vector<double> charges;
    if (!mcc_matches.empty())
    {
      mol.createOrientMap();
      mol.calcOrientMatrices();

      // since OBMol loses charges when adding atoms back the up
      charges.reserve(mol.NumAtoms()+1);
      charges.push_back(0); // move the beginning of the array to 1
      for (int i = 1; i <= mol.NumAtoms(); ++i) charges.push_back(mol.GetAtom(i)->GetPartialCharge());
    }

    mol.BeginModify();
    for (int i = 0; i < mcc_matches.size(); ++i)
    {
      const double eps = 1e-5;
      int idx = mcc_matches[i];
      OpenBabel::OBAtom* a = mol.GetAtom(idx);
      Multipoles *m = static_cast<Multipoles *>(a->GetData("Multipoles"));
      if ( !m ) continue; // no multipoles on this atom, so nothing to do
      if ( ( m->hasDipole() || m->hasQuadrupole() ) )
      {
        double monopole = m->monopole().value();
        Vector3 dipole; if ( m->hasDipole() && (!strip_dipole) ) dipole = m->dipole().value();
        Matrix3 quadrupole; if ( m->hasQuadrupole() && (!strip_quadrupole) ) quadrupole = m->quadrupole().value();

        Matrix3 orient = mol.matrix(idx);
        double rx = radius ;
        double ry = radius ;
        double rz = radius ;
        Eigen::Matrix<double, 7, 7> matrixQ;
        matrixQ <<
        //  charge positions
        //  0       +x             -x             +y             -y             +z             -z
            1,       1,             1,             1,             1,             1,             1,
            0,      rx,           -rx,             0,             0,             0,             0,
            0,       0,             0,            ry,           -ry,             0,             0,
            0,       0,             0,             0,             0,            rz,           -rz,
            0,    rx * rx,       rx * rx,    -.5 * ry * ry, -.5 * ry * ry, -.5 * rz * rz, -.5 * rz * rz,
            0, -.5 * rx * rx, -.5 * rx * rx,    ry * ry,       ry * ry,    -.5 * rz * rz, -.5 * rz * rz,
            0,       0,             0,             0,             0,             0,             0;
        Eigen::Matrix<double, 7, 1> vectorB;
        vectorB << monopole, dipole.x(), dipole.y(), dipole.z(), quadrupole(0,0), quadrupole(1,1), 0.0 ;

        if ( dipole.length() < eps ) // no dipole
        {
          // zeroing charge with smallest component
          double q[3] = {fabs(quadrupole(0,0)), fabs(quadrupole(1,1)), fabs(quadrupole(2,2))};
          if ( q[0] <= q[1] && q[0] <= q[2] ) matrixQ(6,1) = 1; else
          if ( q[1] <= q[0] && q[1] <= q[2] ) matrixQ(6,3) = 1; else
          if ( q[2] <= q[1] && q[2] <= q[0] ) matrixQ(6,5) = 1;
        } else if ( fabs(quadrupole(2,2)) < eps ) // no quadrupole
        {
          // zeroing charge with zero component
          double d[3] = {fabs(dipole(0)), fabs(dipole(1)), fabs(dipole(2))};
          if ( d[0] < eps ) matrixQ(6,1) = 1; else
          if ( d[1] < eps ) matrixQ(6,3) = 1; else
          if ( d[2] < eps ) matrixQ(6,5) = 1; else
            matrixQ(6,0) = 1; // dipole has all components > 0, so zeroing central charge
        } else // both dipole and quadrupole present so dipole controls everything
        {
          // zeroing charge with smallest component
          double d[3] = {fabs(dipole(0)), fabs(dipole(1)), fabs(dipole(2))};
          double q[3] = {fabs(quadrupole(0,0)), fabs(quadrupole(1,1)), fabs(quadrupole(2,2))};
          // check if dipole has all components
          if ( d[0] > eps && d[1] > eps && d[2] > eps ) matrixQ(6,0) = 1; else
            // check if dipole has a single component
          if ( d[0] < eps && d[1] < eps ) // dipole along Z
          if ( q[0] < q[1] ) matrixQ(6,1) = 1; else matrixQ(6,3) = 1; else
          if ( d[0] < eps && d[2] < eps ) // dipole along Y
          if ( q[0] < q[2] ) matrixQ(6,1) = 1; else matrixQ(6,5) = 1; else
          if ( d[1] < eps && d[2] < eps ) // dipole along X
          if ( q[1] < q[2] ) matrixQ(6,3) = 1; else matrixQ(6,5) = 1; else
            // diple has a single zero component
          if ( d[0] < eps ) matrixQ(6,1) = 1; else
          if ( d[1] < eps ) matrixQ(6,3) = 1; else
          if ( d[2] < eps ) matrixQ(6,5) = 1; else
            matrixQ(6,0) = 1; // we should never get here, but just in case ...
        }
        Eigen::Matrix<double, 7, 1> q = matrixQ.fullPivLu().solve(vectorB);

        OpenBabel::OBAtom* central = mol.GetAtom(idx);
        //central->SetPartialCharge(q(0));
        charges[idx] = q(0);
        int vecs[7] = {0, 0, 0, 1, 1, 2, 2};
        double signs[7] = {0, 1, -1, 1, -1, 1, -1};
        for (int j = 1; j < 7; ++j)
        {
          if (fabs(q(j)) > eps)
          {
            OpenBabel::OBAtom *ep = mol.NewAtom();
            OpenBabel::vector3 v =
                signs[i] * OpenBabel::vector3(orient(0,vecs[j]), orient(1,vecs[j]), orient(2,vecs[j])) * radius;
            ep->SetVector(a->GetVector() + v);
            ep->SetAtomicNum(0);
            //ep->SetPartialCharge(q(i));
            charges.push_back(q(j));
            mol.AddBond(idx, ep->GetIdx(), 1);
          }
        }
      }
    }
    mol.EndModify();

    if (!mcc_matches.empty()) for (int i = 1; i <= mol.NumAtoms(); ++i) mol.GetAtom(i)->SetPartialCharge(charges[i]);
    mol.SetPartialChargesPerceived();

    if (!conv.Write(&mol))
    {
      std::cerr << "Error writing final molecule. Aborting." << std::endl;
      return ELEC_TOOLS_OUTPUT_FAILURE;
    }

  } catch (const std::exception& e)
  {
    // generally nothing should fail here but just in case
    std::cerr << "Error: " << e.what() << std::endl;
    return ELEC_TOOLS_UNKNOWN_ERROR;
  }


  return 0;
}
